import shutil

from amino import Either, Boolean, Path, Maybe, L, _, do, IO, Do, Just, Nothing
from amino.boolean import false
from amino.case import Case
from amino.logging import module_log

from ribosome.process import Subprocess

from chromatin.model.venv import VenvStatus, VenvPresent, VenvAbsent, Venv
from chromatin.model.rplugin import (DirRplugin, SiteRplugin, DistRplugin, Rplugin, InstallableRplugin, DistVenvRplugin,
                                     DirVenvRplugin, StackageRplugin, HsDirRplugin, HsStackDirRplugin,
                                     HsStackageRplugin, VenvRplugin, HsInstallableRplugin, HackageRplugin,
                                     HsHackageRplugin)
from chromatin.venv import cons_venv
from chromatin.util.interpreter import python_interpreter

log = module_log()


def venv_present(dir: Path, plugin: Rplugin) -> VenvStatus:
    venv = cons_venv(dir, plugin.name)
    return VenvPresent(plugin, venv)


@do(IO[VenvStatus])
def check_venv(base_dir: Path, plugin: Rplugin) -> Do:
    dir = base_dir / plugin.name
    pip = dir / 'bin' / 'pip'
    exists = yield IO.delay(pip.exists)
    return (
        venv_present(dir, plugin)
        if exists else
        VenvAbsent(plugin)
    )


@do(IO[Boolean])
def venv_exists(base_dir: Path, plugin: Rplugin) -> Do:
    status = yield check_venv(base_dir, plugin)
    return isinstance(status, VenvPresent)


def venv_rplugin_status(base_dir: Path, rplugin: Rplugin) -> IO[VenvStatus]:
    return check_venv(base_dir, rplugin)


class cons_installable_rplugin(Case[Rplugin, Maybe[InstallableRplugin]], alg=Rplugin):

    def dist(self, a: DistRplugin) -> Maybe[InstallableRplugin]:
        return Just(InstallableRplugin(VenvRplugin(DistVenvRplugin(a.spec)), a))

    def dir(self, a: DirRplugin) -> Maybe[InstallableRplugin]:
        return Just(InstallableRplugin(VenvRplugin(DirVenvRplugin(Path(a.spec))), a))

    def site(self, a: SiteRplugin) -> Maybe[InstallableRplugin]:
        return Nothing

    def hackage(self, a: HackageRplugin) -> Maybe[InstallableRplugin]:
        return Just(InstallableRplugin(HsInstallableRplugin(HsHackageRplugin(a.spec)), a))

    def stackage(self, a: StackageRplugin) -> Maybe[InstallableRplugin]:
        return Just(InstallableRplugin(HsInstallableRplugin(HsStackageRplugin(a.spec)), a))

    def hs_dir(self, a: HsDirRplugin) -> Maybe[InstallableRplugin]:
        return Just(InstallableRplugin(HsInstallableRplugin(HsStackDirRplugin(Path(a.spec))), a))


@do(IO[None])
def remove_dir(dir: Path) -> Do:
    exists = yield IO.delay(dir.exists)
    yield IO.delay(shutil.rmtree, dir) if exists else IO.pure(None)


def create_dir(dir: Path) -> IO[None]:
    return IO.delay(dir.mkdir, parents=True, exist_ok=True)


@do(IO[Venv])
def build_venv(global_interpreter: Maybe[str], dir: Path, rplugin_interpreter: Maybe[str], name: str) -> Do:
    interpreter = yield python_interpreter(global_interpreter, rplugin_interpreter)
    retval, out, err = yield Subprocess.popen(str(interpreter), '-m', 'venv', str(dir), '--upgrade', timeout=30)
    success = retval == 0
    yield (
        IO.pure(cons_venv(dir, name))
        if success else
        IO.failed(f'creating venv for `{name}`: {err.join_lines}')
    )


@do(IO[str])
def bootstrap_venv(global_interpreter: Maybe[Path], base_dir: Path, rplugin: Rplugin) -> Do:
    venv_dir = base_dir / rplugin.name
    log.debug(f'bootstrapping {rplugin} in {venv_dir}')
    yield remove_dir(venv_dir)
    yield create_dir(venv_dir)
    yield build_venv(global_interpreter, venv_dir, rplugin.interpreter, rplugin.name)
    return rplugin.name


__all__ = ('check_venv', 'venv_exists', 'venv_package_installed', 'venv_status_check', 'rplugin_ready',
           'venv_rplugin_status', 'cons_installable_rplugin',)



from amino import do, List, _, Lists, Path
from amino.do import Do
from amino.boolean import false, true
from amino.io import IOException
from amino.state import State
from amino.logging import module_log

from ribosome.nvim.io.state import NS
from ribosome.logging import ribo_log
from ribosome.process import SubprocessResult
from ribosome.compute.api import prog
from ribosome.compute.output import Echo
from ribosome.compute.prog import Prog
from ribosome.compute.ribosome_api import Ribo

from chromatin.components.core.logic import (install_plugins, add_installed, reboot_plugins, activate_by_names,
                                             deactivate_by_names, split_plugins_by_install_status)
from chromatin.util import resources
from chromatin.model.venv import Venv
from chromatin.env import Env
from chromatin.settings import venv_dir, autoreboot
from chromatin.components.core.trans.tpe import CrmRibosome

log = module_log()


@prog.echo
@do(NS[Env, Echo])
def install_result(installed: List[SubprocessResult[str]], preinstalled: List[str], errors: List[IOException]) -> Do:
    errors.foreach(lambda e: ribo_log.caught_exception('installing plugin', e))
    success, failed = installed.split(_.success)
    success_rplugins = success / _.data
    failed_rplugins = failed / _.data
    yield (success_rplugins + preinstalled).traverse(add_installed, NS)
    return (
        Echo.error(resources.plugins_install_failed(failed_rplugins))
        if not failed.empty else
        Echo.unit
        if success.empty else
        Echo.info(resources.installed_plugins(success_rplugins))
    )


@prog.echo
@do(NS[Env, Echo])
def installing_message(rplugins: List[str]) -> Do:
    yield NS.pure(
        Echo.none()
        if rplugins.empty else
        Echo.info(resources.installing_plugins(rplugins))
    )


@prog.do(None)
def install_rplugins() -> Do:
    preinstalled, missing = yield split_plugins_by_install_status()
    yield installing_message(missing)
    installed, errors = yield install_plugins(missing, false)
    yield install_result(installed, preinstalled, errors)


@prog.echo
@do(NS[CrmRibosome, Echo])
def activate(*plugins: str) -> Do:
    already_installed = yield activate_by_names(Lists.wrap(plugins))
    names = already_installed / _.name
    return (
        Echo.unit
        if already_installed.empty else
        Echo.info(resources.already_active(names))
    )


@prog
def deactivate(*plugins: str) -> NS[Env, None]:
    return deactivate_by_names(Lists.wrap(plugins))


@do(NS[CrmRibosome, None])
def updated(result: List[str]) -> Do:
    venvs = result.filter_not(lambda a: a == 'chromatin')
    reboot = yield NS.lift(autoreboot.value_or_default())
    if reboot:
        yield reboot_plugins(venvs)


def filter_venvs_by_name(venvs: List[str], names: List[str]) -> List[str]:
    return venvs if names.empty else venvs.filter(lambda v: v in names)


def installed_with_crm() -> State[Env, List[str]]:
    return State.inspect(lambda a: a.ready.cons_m(a.chromatin_venv))


@do(State[Env, List[str]])
def updateable(venv_dir: Path, requested: List[str]) -> Do:
    all = yield installed_with_crm()
    selected = all if requested.empty else all.filter(lambda v: v in requested)


@prog
@do(NS[CrmRibosome, List[str]])
def updateable_venvs(plugins: List[str]) -> Do:
    dir = yield Ribo.setting(venv_dir)
    yield Ribo.zoom_main(updateable(dir, plugins).nvim)


@prog.do(None)
def update_plugins_io(plugins: List[str]) -> Do:
    venvs = yield updateable_venvs(plugins)
    yield (
        Prog.error(resources.no_plugins_match_for_update(plugins))
        if venvs.empty else
        install_plugins(venvs, true)
    )


@prog.echo
@do(NS[CrmRibosome, Echo])
def updated_plugins(results: List[SubprocessResult[Venv]], errors: List[IOException]) -> Do:
    errors.foreach(lambda e: ribo_log.caught_exception('updating plugin', e))
    success, failed = results.split(_.success)
    success_venvs = success / _.data
    failed_venvs = failed / _.data
    yield updated(success_venvs)
    return (
        Echo.info(resources.updated_plugins(success_venvs))
        if failed.empty else
        Echo.error(resources.plugins_install_failed(failed_venvs))
    )


@prog.do(None)
def update_plugins(*ps: str) -> Do:
    plugins = Lists.wrap(ps)
    success, fail = yield update_plugins_io(plugins)
    yield updated_plugins(success, fail)


@prog
def reboot(*plugins: str) -> NS[CrmRibosome, None]:
    return reboot_plugins(Lists.wrap(plugins)).replace(None)


__all__ = ('install_result', 'install_missing', 'update_plugins')



from amino.case import Case
from amino.logging import module_log
from amino import List, do, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.process import Subprocess

from chromatin.model.rplugin import InstallableRpluginMeta, InstallableRplugin, VenvRplugin, HsInstallableRplugin
from chromatin.venv import venv_from_rplugin
from chromatin.install.python import install_venv_rplugin_args
from chromatin.install.haskell import install_hs_rplugin

log = module_log()


class install_rplugin_subproc(Case[InstallableRpluginMeta, NvimIO[Subprocess[str]]], alg=InstallableRpluginMeta):

    def __init__(self, rplugin: InstallableRplugin) -> None:
        self.rplugin = rplugin

    @do(NvimIO[Subprocess[str]])
    def venv(self, venv_rplugin: VenvRplugin) -> Do:
        log.debug(f'installing {venv_rplugin}')
        venv = yield venv_from_rplugin(self.rplugin.rplugin)
        pip_bin = venv.meta.bin_path / 'pip'
        specific_args = install_venv_rplugin_args.match(venv_rplugin.conf)
        extensions = self.rplugin.rplugin.extensions
        args = List('install', '-U', '--no-cache') + specific_args + extensions
        return Subprocess(pip_bin, args, self.rplugin.rplugin.name, timeout=120)

    def hs(self, a: HsInstallableRplugin) -> NvimIO[Subprocess[str]]:
        return install_hs_rplugin(self.rplugin)(a.conf)


__all__ = ('install_rplugin_subproc',)



import os

from amino import Path, do, Do

from ribosome.nvim.io.api import N
from ribosome.process import Subprocess
from ribosome.nvim.io.compute import NvimIO

from chromatin.util.interpreter import stack_exe
from chromatin.model.rplugin import ActiveRplugin, Rplugin
from chromatin.activate.host import start_rplugin_host
from chromatin.host import start_host


@do(NvimIO[ActiveRplugin])
def activate_stack_plugin(rplugin: Rplugin, dir: Path) -> Do:
    stack = yield N.from_io(stack_exe())
    code, out, err = yield N.from_io(Subprocess.popen(stack, 'path', '--local-bin', cwd=dir, env=None))
    bin_path = yield N.m(out.head, lambda: f'error running `stack path --local-bin: {err.join_lines}`')
    exe = Path(bin_path) / rplugin.name
    yield start_rplugin_host(rplugin, lambda debug: start_host(str(exe), debug))


def cabal_rplugin_executable(rplugin: Rplugin) -> Path:
    return Path.home() / '.cabal' / 'bin' / rplugin.name


def cabal_rplugin_cmdline(exe: Path, name: str, debug: bool) -> str:
    extra = (
        f'-l /tmp/chromatin-{name}-{os.getpid()} -v DEBUG'
        if debug else
        ''
    )
    return f'{exe} {extra}'


@do(NvimIO[ActiveRplugin])
def activate_cabal_plugin(rplugin: Rplugin, dir: Path) -> Do:
    exe = cabal_rplugin_executable(rplugin)
    yield start_rplugin_host(rplugin, lambda debug: start_host(cabal_rplugin_cmdline(exe, rplugin.name, debug), debug))


__all__ = ('activate_stack_plugin', 'activate_cabal_plugin', 'cabal_rplugin_executable',)



from ribosome.compute.ribosome import Ribosome
from ribosome.config.basic_config import NoData

from chromatin.config.component import ChromatinComponent
from chromatin.env import Env

CrmRibosome = Ribosome[Env, ChromatinComponent, NoData]

__all__ = ('CrmRibosome',)



from amino import List, Just

from ribosome.rpc.api import rpc
from ribosome.config.config import Config
from ribosome.rpc.data.prefix_style import Plain, Full

from chromatin.env import Env
from chromatin.components.core.trans.install import update_plugins, activate, deactivate, reboot
from chromatin.components.core.trans.setup import init, setup_plugins, show_plugins, add_plugin

chromatin_config: Config = Config.cons(
    name='chromatin',
    prefix='crm',
    state_ctor=Env.cons,
    rpc=List(
        rpc.write(init).conf(prefix=Full()),
        rpc.write(add_plugin).conf(name=Just('cram'), prefix=Plain(), json=True),
        rpc.write(setup_plugins),
        rpc.write(show_plugins),
        rpc.write(activate),
        rpc.write(deactivate),
        rpc.write(reboot),
        rpc.write(update_plugins).conf(name=Just('update')),
    ),
    init=init,
)

__all__ = ('config',)



from typing import Tuple, Callable

from amino import do, Just, List, Either, Nil, Boolean, Path, Lists, IO
from amino.do import Do
from amino.state import State
from amino.io import IOException
from amino.func import ReplaceVal
from amino.util.string import camelcaseify, camelcase
from amino.case import Case
from amino.logging import module_log
from amino.list import split_by_status_zipped, split_either_list

from ribosome.nvim.io.state import NS
from ribosome.process import SubprocessResult, Subprocess
from ribosome.nvim.io.compute import NvimIO
from ribosome.compute.api import prog
from ribosome.compute.output import (GatherSubprocesses, GatherIOResult, GatherItem, GatherResult,
                                     GatherSubprocessResult, GatherIO, GatherSubprocess, Gather)
from ribosome.nvim.io.api import N
from ribosome.nvim.api.command import runtime, nvim_command
from ribosome.nvim.api.exists import command_exists, wait_until_function_produces, wait_for_function, function_exists
from ribosome.nvim.api.function import nvim_call_json, nvim_call_function
from ribosome.rpc.define import ActiveRpcTrigger, undef_command
from ribosome.compute.ribosome_api import Ribo
from ribosome.components.internal.prog import RpcTrigger
from ribosome import ribo_log

from chromatin.host import stop_host
from chromatin.util import resources
from chromatin.model.rplugin import (Rplugin, ActiveRplugin, DirRplugin, DistRplugin, SiteRplugin, InstallableRplugin,
                                     InstallableRpluginMeta, DistVenvRplugin, DirVenvRplugin, StackageRplugin,
                                     HsDirRplugin, VenvRplugin, HsInstallableRplugin, HsRpluginMeta, VenvRpluginMeta,
                                     HsStackageRplugin, HsStackDirRplugin, HackageRplugin, HsHackageRplugin)
from chromatin.env import Env
from chromatin.components.core.trans.tpe import CrmRibosome
from chromatin.settings import handle_crm, venv_dir, rplugins
from chromatin.util.interpreter import join_pythonpath
from chromatin.venv import cons_venv, venv_plugin_path, venv_package_installed, venv_from_rplugin
from chromatin.model.venv import Venv
from chromatin.components.core.rplugin import installable_rplugin_from_name, installable_rplugins_from_names
from chromatin.install.main import install_rplugin_subproc
from chromatin.activate.haskell import activate_stack_plugin, activate_cabal_plugin, cabal_rplugin_executable
from chromatin.activate.host import start_rplugin_host, start_python_rplugin_host

log = module_log()


@do(NvimIO[List[ActiveRpcTrigger]])
def define_handlers(active_rplugin: ActiveRplugin) -> Do:
    channel = active_rplugin.channel
    cname = camelcaseify(active_rplugin.name)
    rpc_handlers_fun = f'{cname}RpcHandlers'
    result = yield N.call_once_defined(rpc_handlers_fun, timeout=3)
    handlers = (
        Lists.wrap(result)
        .flat_map(RpcHandlerSpec.decode)
        .map(lambda spec: ActiveRpcTrigger(spec=spec, channel=channel))
    )
    yield N.pure(handlers)


# FIXME handlers cannot be obtained while the plugin is initializing, as the RpcHandlers function will be reported as
# 'in use' when it is being polled during definition
# fetch handlers when using them, i.e. during shutdown
@do(NS[Env, None])
def activated(active_rplugin: ActiveRplugin) -> Do:
    log.debug(f'activated {active_rplugin}')
    spec = active_rplugin.rplugin
    yield NS.lift(runtime(f'chromatin/{spec.name}/*'))
    yield NS.modify(lambda a: a.host_started(active_rplugin.meta))


class activate_rplugin_io(Case[Rplugin, NvimIO[ActiveRplugin]], alg=Rplugin):

    @do(NvimIO[ActiveRplugin])
    def dir_rplugin(self, rplugin: DirRplugin) -> Do:
        venv = yield venv_from_rplugin(rplugin)
        python_exe = venv.meta.python_executable
        bin_path = venv.meta.bin_path
        plugin_path = Path(rplugin.spec) / rplugin.name / '__init__.py'
        yield start_python_rplugin_host(rplugin, python_exe, bin_path, plugin_path, List(rplugin.spec))

    @do(NvimIO[ActiveRplugin])
    def dist_rplugin(self, rplugin: DistRplugin) -> Do:
        venv = yield venv_from_rplugin(rplugin)
        plugin_path_e = yield N.from_io(venv_plugin_path(venv))
        plugin_path = yield N.e(plugin_path_e)
        python_exe = venv.meta.python_executable
        bin_path = venv.meta.bin_path
        yield start_python_rplugin_host(rplugin, python_exe, bin_path, plugin_path, Nil)

    # TODO start all plugins by passing the module path to ribosome_start_plugin instead of __init__.py
    def site_rplugin(self, rplugin: SiteRplugin) -> NvimIO[ActiveRplugin]:
        return N.error('site rplugins not implemented yet')

    def hackage(self, rplugin: HackageRplugin) -> NvimIO[ActiveRplugin]:
        return activate_cabal_plugin(rplugin, Path.home())

    def stackage(self, rplugin: StackageRplugin) -> NvimIO[ActiveRplugin]:
        return activate_stack_plugin(rplugin, Path.home())

    def hs_dir_rplugin(self, rplugin: HsDirRplugin) -> NvimIO[ActiveRplugin]:
        return activate_stack_plugin(rplugin, rplugin.spec)


@do(NS[CrmRibosome, None])
def activate_rplugin(rplugin: Rplugin) -> Do:
    active_rplugin = yield NS.lift(activate_rplugin_io.match(rplugin))
    yield Ribo.zoom_main(activated(active_rplugin))


@do(NS[CrmRibosome, List[Rplugin]])
def activate_multi(new_rplugins: List[Rplugin]) -> Do:
    active = yield Ribo.zoom_main(NS.inspect_either(lambda a: a.active_rplugins))
    active_rplugins = active.map(lambda a: a.rplugin)
    already_active, inactive = new_rplugins.split(active_rplugins.contains)
    yield inactive.traverse(activate_rplugin, NS).replace(already_active)


@do(NS[CrmRibosome, List[Rplugin]])
def activate_by_names(plugins: List[str]) -> Do:
    getter = (lambda a: a.ready_rplugins) if plugins.empty else (lambda a: a.ready_by_name(plugins))
    rplugins = yield Ribo.zoom_main(NS.inspect_either(getter))
    yield (
        NS.error(resources.no_plugins_match_for_activation(plugins))
        if rplugins.empty else
        activate_multi(rplugins)
    )


def activate_all() -> NS[CrmRibosome, List[Rplugin]]:
    return activate_by_names(Nil)


@do(NvimIO[None])
def undef_trigger(trigger: RpcTrigger) -> Do:
    method = undef_command.match(trigger.method)
    yield method.map(lambda a: nvim_command(a, trigger.name)).get_or_strict(N.unit)


@do(NvimIO[None])
def stop_rplugin(name: str, channel: int, triggers: List[RpcTrigger]) -> Do:
    yield nvim_command(f'{camelcase(name)}Quit')
    yield triggers.traverse(undef_trigger, NvimIO)
    yield nvim_command('autocmd!', name)
    yield stop_host(channel)


@do(NS[Env, None])
def deactivate_rplugin(active_rplugin: ActiveRplugin) -> Do:
    rplugin = active_rplugin.rplugin
    meta = active_rplugin.meta
    cname = camelcaseify(rplugin.name)
    rpc_triggers_fun = f'{cname}RpcTriggers'
    triggers = yield NS.lift(nvim_call_json(rpc_triggers_fun))
    yield NS.modify(lambda a: a.deactivate_rplugin(meta))
    yield NS.lift(stop_rplugin(rplugin.name, meta.channel, triggers))
    log.debug(f'deactivated {active_rplugin}')
    yield NS.pure(None)


def deactivate_multi(rplugins: List[ActiveRplugin]) -> NS[Env, None]:
    return rplugins.traverse(deactivate_rplugin, NS).replace(None)


@do(NS[Env, None])
def deactivate_by_names(plugins: List[str]) -> Do:
    getter = (lambda a: a.active_rplugins) if plugins.empty else (lambda a: a.active_rplugins_by_name(plugins))
    rplugins = yield NS.inspect_either(getter)
    yield (
        NS.error(resources.no_plugins_match_for_deactivation(plugins))
        if rplugins.empty else
        deactivate_multi(rplugins)
    )


@do(NS[CrmRibosome, List[Rplugin]])
def reboot_plugins(plugins: List[str]) -> Do:
    yield Ribo.zoom_main(deactivate_by_names(plugins))
    yield activate_by_names(plugins)


@do(NvimIO[None])
def rplugin_stage(prefix: str, num: int) -> Do:
    name = f'{prefix}Stage{num}'
    exists = yield command_exists(name)
    if exists:
        yield nvim_command(name)
    else:
        fun_exists = yield function_exists(name)
        if fun_exists:
            yield nvim_call_function(name)


def init_stage(prefixes: List[str]) -> Callable[[int], NvimIO[None]]:
    def init_stage(num: int) -> NvimIO[None]:
        return prefixes.traverse(lambda a: rplugin_stage(a, num), NvimIO)
    return init_stage


@do(NvimIO[None])
def await_poll(prefix: str) -> Do:
    yield wait_for_function(f'{prefix}Poll', 10)
    yield wait_until_function_produces(True, f'{prefix}Poll')


@do(NvimIO[List[str]])
def initialize_plugins(rplugins: List[ActiveRplugin]) -> Do:
    prefixes = rplugins.map(lambda a: camelcase(a.name))
    yield prefixes.traverse(await_poll, NvimIO)
    yield Lists.range(1, 6).traverse(init_stage(prefixes), NvimIO)
    return Nil


# TODO don't fail if one plugin doesn't start
@do(NS[Env, None])
def activation_complete() -> Do:
    rplugins = yield NS.inspect_either(lambda a: a.uninitialized_rplugins)
    tracked = rplugins.filter(lambda a: a.rplugin.track)
    failed = yield NS.lift(initialize_plugins(tracked))
    yield NS.modify(lambda a: a.initialization_complete(failed))


@do(NS[CrmRibosome, None])
def activate_newly_installed() -> Do:
    new = yield Ribo.zoom_main(NS.inspect_either(lambda a: a.inactive))
    yield activate_multi(new)
    yield Ribo.zoom_main(activation_complete())


def add_venv(name: str) -> State[Env, None]:
    return State.modify(lambda s: s if name in s.venvs else s.append1.venvs(name))


@do(State[Env, None])
def add_installed(rplugin: Rplugin) -> Do:
    yield State.modify(lambda s: s.copy(ready=s.ready.cat(rplugin).distinct))


@do(NS[Env, None])
def already_installed(venv_dir: Path, venv_rplugin: InstallableRplugin) -> Do:
    yield add_installed(venv_rplugin.rplugin).nvim
    venv = cons_venv(venv_dir, venv_rplugin.rplugin.name)
    yield add_venv(venv.meta).nvim


@prog.subproc.gather
@do(NS[Env, GatherSubprocesses[Venv]])
def install_plugins_procs(names: List[str], update: Boolean) -> Do:
    action = 'updating' if update else 'installing'
    log.debug(f'{action} rplugins: {names}')
    rplugins = yield installable_rplugins_from_names(names)
    procs = yield NS.lift(rplugins.traverse(lambda a: install_rplugin_subproc(a)(a.meta), NvimIO))
    return GatherSubprocesses(procs, timeout=600)


@prog
def install_plugins_result(results: List[Either[IOException, SubprocessResult[str]]]
                           ) -> NS[Env, Tuple[List[SubprocessResult[str]], List[str]]]:
    venvs = results.flat_map(lambda a: a.cata(ReplaceVal(Nil), List))
    errors = results.flat_map(lambda a: a.cata(List, ReplaceVal(Nil)))
    log.debug(f'installed plugins: {venvs}')
    if errors:
        log.debug(f'error installing plugins: {errors}')
    return NS.pure((venvs, errors))


@prog.do(Tuple[List[SubprocessResult[str]], List[str]])
def install_plugins(names: List[str], update: Boolean) -> Do:
    result = yield install_plugins_procs(names, update)
    yield install_plugins_result(result)


@do(NS[CrmRibosome, None])
def add_crm_venv() -> Do:
    handle = yield Ribo.setting(handle_crm)
    if handle:
        log.debug('adding chromatin venv')
        plugin = Rplugin.simple('chromatin')
        yield Ribo.modify_main(lambda a: a.set.chromatin_rplugin(Just(plugin)))
        yield Ribo.modify_main(lambda a: a.set.chromatin_venv(Just('chromatin')))


@do(NS[Env, List[Rplugin]])
def read_conf() -> Do:
    plugin_conf = yield Ribo.setting(rplugins)
    specs = plugin_conf.traverse(Rplugin.from_config, Either)
    yield NS.from_either(specs)


class venv_rplugin_healthy(
        Case[VenvRpluginMeta, NS[CrmRibosome, GatherItem[Tuple[str, bool]]]],
        alg=VenvRpluginMeta,
):

    def __init__(self, rplugin: Rplugin) -> None:
        self.rplugin = rplugin

    @do(NS[CrmRibosome, GatherItem[Tuple[str, bool]]])
    def dist(self, rplugin: DistVenvRplugin) -> Do:
        venv = yield NS.lift(venv_from_rplugin(self.rplugin))
        return GatherIO(venv_package_installed(venv).map(lambda status: (self.rplugin.name, status)))

    @do(NS[CrmRibosome, GatherItem[Tuple[str, bool]]])
    def dir(self, rplugin: DirVenvRplugin) -> Do:
        venv = yield NS.lift(venv_from_rplugin(self.rplugin))
        args = List('-c', f'import {self.rplugin.name}')
        path = join_pythonpath(self.rplugin.pythonpath.cons(rplugin.dir))
        return GatherSubprocess(
            Subprocess(str(venv.meta.python_executable), args, self.rplugin.name, timeout=1, env=dict(PYTHONPATH=path))
        )


class hs_rplugin_healthy(
        Case[HsRpluginMeta, NS[CrmRibosome, GatherItem[Tuple[str, bool]]]],
        alg=HsRpluginMeta,
):

    def __init__(self, rplugin: Rplugin) -> None:
        self.rplugin = rplugin

    def hackage(self, a: HsHackageRplugin) -> NS[CrmRibosome, GatherItem[Tuple[str, bool]]]:
        exe = cabal_rplugin_executable(self.rplugin)
        return NS.pure(GatherIO(IO.pure((self.rplugin.name, exe.is_file()))))

    def stackage(self, a: HsStackageRplugin) -> NS[CrmRibosome, GatherItem[Tuple[str, bool]]]:
        return NS.pure(GatherIO(IO.pure((self.rplugin.name, False))))

    def dir(self, a: HsStackDirRplugin) -> NS[CrmRibosome, GatherItem[Tuple[str, bool]]]:
        return NS.pure(GatherIO(IO.pure((self.rplugin.name, False))))



class rplugin_healthy(
        Case[InstallableRpluginMeta, NS[CrmRibosome, GatherItem[Tuple[str, bool]]]],
        alg=InstallableRpluginMeta,
):

    def __init__(self, rplugin: Rplugin) -> None:
        self.rplugin = rplugin

    def venv(self, meta: VenvRplugin) -> NS[CrmRibosome, GatherItem[Tuple[str, bool]]]:
        return venv_rplugin_healthy(self.rplugin)(meta)

    def hs(self, meta: HsInstallableRplugin) -> NS[CrmRibosome, GatherItem[Tuple[str, bool]]]:
        return hs_rplugin_healthy(self.rplugin)(meta)


class healthcheck_result(
        Case[GatherResult[Tuple[InstallableRplugin, bool]], Tuple[InstallableRplugin, bool]],
        alg=GatherResult,
):

    def io(self, a: GatherIOResult[Tuple[InstallableRplugin, bool]]) -> Tuple[InstallableRplugin, bool]:
        return a.result

    def subprocess(self, a: GatherSubprocessResult[Tuple[InstallableRplugin, bool]]) -> Tuple[InstallableRplugin, bool]:
        return a.result.data, a.result.retval == 0


@prog.gather
@do(NS[CrmRibosome, List[Either[IOException, GatherIOResult[Tuple[InstallableRplugin, bool]]]]])
def health_statuses() -> Do:
    dir = yield Ribo.setting(venv_dir)
    all_venvs = yield Ribo.inspect_main(lambda a: a.venvs)
    log.debug(f'running health check on venvs: {all_venvs.join_comma}')
    installable = yield Ribo.zoom_main(all_venvs.traverse(lambda a: installable_rplugin_from_name(dir, a), NS))
    gather_items = yield installable.traverse(lambda a: rplugin_healthy(a.rplugin)(a.meta), NS)
    return Gather(gather_items, 5)


@prog.do(Tuple[List[str], List[str]])
def split_plugins_by_install_status() -> Do:
    output = yield health_statuses()
    fatal, success = split_either_list(output)
    if fatal:
        ribo_log.error(f'failed to run healthcheck for plugins: {fatal.join_comma}')
    statuses = success.map(healthcheck_result.match)
    present, absent = split_by_status_zipped(statuses)
    return present, absent


def store_errors(errors: List[str]) -> Callable[[Env], Env]:
    return lambda e: e.append.errors(errors)


__all__ = ('add_crm_venv', 'read_conf', 'install_plugins', 'add_installed', 'split_plugins_by_install_status',
           'store_errors',)



from typing import Callable, Tuple

from amino import do, Do, Path, List

from ribosome.nvim.io.compute import NvimIO

from chromatin.model.rplugin import ActiveRplugin, Rplugin, ActiveRpluginMeta
from chromatin.settings import debug_pythonpath
from chromatin.host import start_python_host


@do(NvimIO[ActiveRplugin])
def start_rplugin_host(rplugin: Rplugin, start: Callable[[bool], NvimIO[Tuple[int, int]]]) -> Do:
    debug_global = yield debug_pythonpath.value
    debug = debug_global.get_or_strict(rplugin.debug)
    channel, pid = yield start(debug)
    return ActiveRplugin(rplugin, ActiveRpluginMeta(rplugin.name, channel, pid))


@do(NvimIO[ActiveRplugin])
def start_python_rplugin_host(
        rplugin: Rplugin,
        python_exe: Path,
        bin_path: Path,
        plugin_path: Path,
        pythonpath: List[str],
) -> Do:
    yield start_rplugin_host(
        rplugin,
        lambda debug: start_python_host(python_exe, bin_path, plugin_path, debug, pythonpath + rplugin.pythonpath),
    )

__all__ = ('start_rplugin_host', 'start_python_rplugin_host',)



import pkg_resources

from amino import Path, IO, do, Boolean, Maybe, Lists, Either
from amino.boolean import false
from amino.do import Do
from amino.logging import module_log
from amino.case import Case

from ribosome.nvim.io.compute import NvimIO

from chromatin.util.interpreter import python_interpreter
from chromatin.model.venv import (Venv, VenvMeta, VenvPackageExistent, VenvPackageAbsent, VenvPackageStatus, VenvStatus,
                                  VenvPresent, VenvAbsent)
from chromatin.model.rplugin import Rplugin
from chromatin.settings import venv_dir

log = module_log()


@do(IO[Either[str, Path]])
def venv_site(venv: Venv) -> Do:
    lib_dir = venv.meta.dir / 'lib'
    libs = yield IO.delay(lib_dir.glob, 'python3.?')
    lib = Lists.wrap(libs).head.to_either(f'no python dirs in {lib_dir}')
    return lib.map(lambda a: a / 'site-packages')


@do(IO[Either[str, Path]])
def venv_plugin_path(venv: Venv) -> Do:
    site = yield venv_site(venv)
    return site.map(lambda a: a / venv.name / '__init__.py')


def cons_venv(dir: Path, name: str) -> Venv:
    bin_path = dir / 'bin'
    executable = bin_path / 'python'
    return Venv(name, VenvMeta(name, dir, executable, bin_path))


def cons_venv_under(base_dir: Path, name: str) -> Venv:
    return cons_venv(base_dir / name, name)


@do(IO[VenvPackageStatus])
def venv_package_status_site(venv: Venv, site: Path, req_spec: str) -> Do:
    ws = yield IO.delay(pkg_resources.WorkingSet, [site])
    req = yield IO.delay(pkg_resources.Requirement, req_spec)
    return Maybe.check(ws.by_key.get(req.key)).cata_strict(
        lambda a: VenvPackageExistent(venv, a),
        VenvPackageAbsent(venv),
    )


@do(IO[VenvPackageStatus])
def venv_package_status(venv: Venv, req: str) -> Do:
    site_e = yield venv_site(venv)
    yield site_e.cata(lambda a: IO.pure(VenvPackageAbsent(venv)), lambda a: venv_package_status_site(venv, a, req))


@do(IO[Boolean])
def venv_package_installed(venv: Venv) -> Do:
    status = yield venv_package_status(venv, venv.name)
    return status.exists


class venv_status_check(Case[VenvStatus, IO[Boolean]], alg=VenvStatus):

    def venv_present(self, status: VenvPresent) -> IO[Boolean]:
        return venv_package_installed(status.venv)

    def venv_absent(self, status: VenvAbsent) -> IO[Boolean]:
        return IO.pure(false)


@do(NvimIO[Venv])
def venv_from_rplugin(rplugin: Rplugin) -> Do:
    dir = yield venv_dir.value_or_default()
    return cons_venv_under(dir, rplugin.name)


__all__ = ('venv_site', 'venv_plugin_path', 'cons_venv', 'cons_venv_under', 'venv_package_status_site',
           'venv_package_status', 'venv_package_installed', 'venv_status_check', 'venv_from_rplugin',)



import typing
from typing import Tuple

from amino import Path, do, List, Nil
from amino.do import Do
from amino.logging import module_log

from ribosome.nvim.io.compute import NvimIO
from ribosome.logging import ribo_log
from ribosome.nvim.api.function import nvim_call_function, nvim_call_tpe
from ribosome.nvim.io.api import N
from ribosome.rpc.error import define_rpc_stderr_handler

log = module_log()
stderr_handler_prefix = 'Chromatin'


def python_host_cmdline(
        python_exe: Path,
        bin_path: Path,
        plug: Path,
        debug: bool,
        pythonpath: List[str],
) -> typing.List[str]:
    debug_option = [] if debug else ['-E']
    ppath = pythonpath.mk_string(':')
    pre = [] if pythonpath.empty else ['env', f'RIBOSOME_PYTHONPATH={ppath}']
    args = [str(bin_path / f'ribosome_start_plugin'), str(plug)]
    return pre + [str(python_exe)] + debug_option + args


@do(NvimIO[Tuple[int, int]])
def start_host(
        cmdline: str,
        debug: bool=False,
) -> Do:
    ribo_log.debug(f'starting host: {cmdline}; debug: {debug}')
    stderr_handler_name = yield define_rpc_stderr_handler(stderr_handler_prefix)
    channel = yield nvim_call_tpe(int, 'jobstart', cmdline, dict(rpc=True, on_stderr=stderr_handler_name))
    pid = yield nvim_call_tpe(int, 'jobpid', channel)
    ribo_log.debug(f'host running, channel {channel}, pid {pid}')
    yield N.pure((channel, pid))


@do(NvimIO[Tuple[int, int]])
def start_python_host(
        python_exe: Path,
        bin_path: Path,
        plugin_path: Path,
        debug: bool=False,
        pythonpath: List[str]=Nil,
) -> Do:
    cmdline = python_host_cmdline(python_exe, bin_path, plugin_path, debug, pythonpath)
    yield start_host(cmdline, debug)


@do(NvimIO[None])
def stop_host(channel: int) -> Do:
    yield nvim_call_function('jobstop', channel)


__all__ = ('start_python_host', 'stop_host', 'start_host',)



from typing import Type, Any, Tuple
import json

from amino import Either, Map, Regex, do, Do, Try, List, Maybe, Nil, Nothing, Path
from amino.dat import ADT, Dat
from amino.json.decoder import decode_json_type


class ConfigRplugin(Dat['ConfigRplugin']):

    def __init__(
            self,
            spec: str,
            name: Maybe[str],
            debug: Maybe[bool],
            pythonpath: Maybe[List[str]],
            interpreter: Maybe[str],
            extensions: Maybe[List[str]],
            track: Maybe[bool],
    ) -> None:
        self.spec = spec
        self.name = name
        self.debug = debug
        self.pythonpath = pythonpath
        self.interpreter = interpreter
        self.extensions = extensions
        self.track = track


class Rplugin(ADT['Rplugin']):

    @classmethod
    def cons(
            cls,
            name: str,
            spec: str,
            debug: bool=False,
            pythonpath: List[str]=Nil,
            interpreter: Maybe[str]=Nothing,
            extensions: List[str]=Nil,
            track: bool=True,
    ) -> 'Rplugin':
        return cls(name, spec, debug, pythonpath, interpreter, extensions, track)

    def __init__(
            self,
            name: str,
            spec: str,
            debug: bool,
            pythonpath: List[str],
            interpreter: Maybe[str],
            extensions: List[str],
            track: bool,
    ) -> None:
        self.name = name
        self.spec = spec
        self.debug = debug
        self.pythonpath = pythonpath
        self.interpreter = interpreter
        self.extensions = extensions
        self.track = track

    @property
    def pythonpath_str(self) -> None:
        return self.pythonpath.mk_string(':')

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.name == other.name and self.spec == other.spec

    @staticmethod
    @do(Either[str, 'Rplugin'])
    def from_config(data: dict) -> Do:
        json_conf = yield Try(json.dumps, data)
        conf = yield decode_json_type(json_conf, ConfigRplugin)
        return cons_rplugin(conf)

    @staticmethod
    def simple(name: str) -> 'Rplugin':
        return DistRplugin.cons(name=name, spec=name)


class DistRplugin(Rplugin):
    pass


class DirRplugin(Rplugin):
    pass


class SiteRplugin(Rplugin):
    pass


class HackageRplugin(Rplugin):
    pass


class StackageRplugin(Rplugin):
    pass


class HsDirRplugin(Rplugin):
    pass


ctors: Map[str, Type[Rplugin]] = Map(
    dir=DirRplugin,
    site=SiteRplugin,
    venv=DistRplugin,
    hackage=HackageRplugin,
    stackage=StackageRplugin,
    hs_dir=HsDirRplugin,
)
prefixes = ctors.k.mk_string('|')
spec_rex = Regex(f'(?P<prefix>{prefixes}):(?P<spec>.*)')


@do(Either[str, Tuple[Type[Rplugin], str]])
def parse_spec(raw_spec: str) -> Do:
    match = yield spec_rex.match(raw_spec)
    prefix, spec = yield match.all_groups('prefix', 'spec')
    tpe = yield ctors.lift(prefix).to_either('invalid rplugin spec prefix `{prefix}`')
    return tpe, spec


def cons_rplugin(conf: ConfigRplugin) -> Rplugin:
    tpe, spec = parse_spec(conf.spec).get_or_strict((DistRplugin, conf.spec))
    return tpe.cons(
        conf.name.get_or_strict(spec),
        spec,
        conf.debug.get_or_strict(False),
        conf.pythonpath.get_or_strict(Nil),
        conf.interpreter,
        conf.extensions.get_or_strict(Nil),
        conf.track.get_or_strict(True),
    )


class HsRpluginMeta(ADT['HsRpluginMeta']):
    pass


class HsStackDirRplugin(HsRpluginMeta):

    def __init__(self, dir: Path) -> None:
        self.dir = dir


class HsHackageRplugin(HsRpluginMeta):

    def __init__(self, dep: str) -> None:
        self.dep = dep


class HsStackageRplugin(HsRpluginMeta):

    def __init__(self, dep: str) -> None:
        self.dep = dep


class VenvRpluginMeta(ADT['VenvRpluginMeta']):
    pass


class DistVenvRplugin(VenvRpluginMeta):

    def __init__(self, req: str) -> None:
        self.req = req


class DirVenvRplugin(VenvRpluginMeta):

    def __init__(self, dir: Path) -> None:
        self.dir = dir


class InstallableRpluginMeta(ADT['InstallableRplugin']):
    pass


class VenvRplugin(InstallableRpluginMeta):

    def __init__(self, conf: VenvRpluginMeta) -> None:
        self.conf = conf


class HsInstallableRplugin(InstallableRpluginMeta):

    def __init__(self, conf: HsRpluginMeta) -> None:
        self.conf = conf


class InstallableRplugin(Dat['InstallableRplugin']):

    def __init__(self, meta: InstallableRpluginMeta, rplugin: Rplugin) -> None:
        self.meta = meta
        self.rplugin = rplugin


class ActiveRpluginMeta(Dat['ActiveRpluginMeta']):

    def __init__(self, rplugin: str, channel: int, pid: int) -> None:
        self.rplugin = rplugin
        self.channel = channel
        self.pid = pid


class ActiveRplugin(Dat['ActiveRplugin']):

    def __init__(self, rplugin: Rplugin, meta: ActiveRpluginMeta) -> None:
        self.rplugin = rplugin
        self.meta = meta

    @property
    def name(self) -> str:
        return self.rplugin.name


__all__ = ('Rplugin', 'DistRplugin', 'DirRplugin', 'SiteRplugin', 'cons_rplugin', 'ActiveRpluginMeta', 'ActiveRplugin',
           'InstallableRplugin', 'DistVenvRplugin', 'DirVenvRplugin', 'StackageRplugin', 'HsDirRplugin',
           'InstallableRpluginMeta', 'HsHackageRplugin',)



from typing import Tuple

from amino import do, List, Boolean, Either, Nil, IO, Maybe, Dat, Path
from amino.do import Do
from amino.logging import module_log
from amino.case import Case
from amino.state import State
from ribosome.nvim.io.state import NS
from ribosome import ribo_log
from ribosome.compute.api import prog
from ribosome.compute.output import Echo, GatherIOs
from ribosome.compute.ribosome_api import Ribo

from chromatin.components.core.logic import add_crm_venv, read_conf, activate_newly_installed, add_venv, store_errors
from chromatin.model.rplugin import (Rplugin, cons_rplugin, ConfigRplugin, InstallableRplugin, InstallableRpluginMeta,
                                     VenvRplugin, HsInstallableRplugin)
from chromatin.components.core.trans.install import install_rplugins
from chromatin.model.venv import Venv, VenvStatus, VenvPresent, VenvAbsent
from chromatin.util import resources
from chromatin.settings import venv_dir, autostart, interpreter
from chromatin.env import Env
from chromatin.components.core.trans.tpe import CrmRibosome
from chromatin.rplugin import cons_installable_rplugin, bootstrap_venv, venv_rplugin_status

log = module_log()


@prog
@do(NS[CrmRibosome, List[Rplugin]])
def initialize() -> Do:
    yield add_crm_venv()
    yield Ribo.zoom_main(read_conf())


@prog
@do(NS[Env, None])
def bootstrap_result(result: List[Either[str, str]]) -> Do:
    def split(z: Tuple[List[str], List[Venv]], a: Either[str, Venv]) -> Tuple[List[str], List[Venv]]:
        err, vs = z
        return a.cata((lambda e: (err.cat(e), vs)), (lambda v: (err, vs.cat(v))))
    errors, venvs = result.fold_left((Nil, Nil))(split)
    ribo_log.debug(f'bootstrapped venvs: {venvs}')
    yield venvs.traverse(add_venv, State).nvim
    yield NS.modify(store_errors(errors))
    yield NS.error(f'failed to setup venvs: {errors.map(str).join_comma}') if errors else NS.unit


class bootstrap_venv_rplugin(Case[VenvStatus, IO[str]], alg=VenvStatus):

    def __init__(self, global_interpreter: Maybe[Path], dir: Path) -> None:
        self.global_interpreter = global_interpreter
        self.dir = dir

    def present(self, status: VenvPresent) -> IO[str]:
        return IO.pure(status.venv.name)

    def absent(self, status: VenvAbsent) -> IO[str]:
        return bootstrap_venv(self.global_interpreter, self.dir, status.rplugin)


@do(NS[CrmRibosome, List[InstallableRplugin]])
def installable_plugins() -> Do:
    plugins = yield Ribo.inspect_main(lambda a: a.rplugins)
    return plugins.flat_map(cons_installable_rplugin.match)


@do(NS[CrmRibosome, List[IO[str]]])
def bootstrap_hs_rplugin(rplugin: InstallableRplugin, meta: HsInstallableRplugin) -> None:
    yield NS.pure(IO.pure(rplugin.rplugin.name))


class bootstrap_rplugin(
        Case[InstallableRpluginMeta, NS[CrmRibosome, List[IO[List[Either[str, str]]]]]],
        alg=InstallableRpluginMeta,
):

    def __init__(self, rplugin: InstallableRplugin, dir: Path, global_interpreter: Maybe[Path]) -> None:
        self.rplugin = rplugin
        self.dir = dir
        self.global_interpreter = global_interpreter

    @do(NS[CrmRibosome, List[IO[str]]])
    def venv(self, venv_rplugin: VenvRplugin) -> Do:
        status = yield NS.from_io(venv_rplugin_status(self.dir, self.rplugin.rplugin))
        return bootstrap_venv_rplugin(self.global_interpreter, self.dir)(status)

    def hs(self, a: HsInstallableRplugin) -> NS[CrmRibosome, List[IO[str]]]:
        return bootstrap_hs_rplugin(self.rplugin, a)


@prog.io.gather
@do(NS[CrmRibosome, GatherIOs[str]])
def bootstrap_rplugins_io() -> Do:
    vr = yield installable_plugins()
    dir = yield Ribo.setting(venv_dir)
    global_interpreter = yield Ribo.setting_raw(interpreter)
    ios = yield vr.traverse(lambda a: bootstrap_rplugin(a, dir, global_interpreter.to_maybe)(a.meta), NS)
    yield NS.pure(GatherIOs(ios, timeout=30))


@prog.do(None)
def bootstrap_rplugins() -> Do:
    '''check whether a venv exists for each venv or dir plugin in the env.
    for those that don't have one, create venvs in `g:chromatin_venv_dir`.
    '''
    result = yield bootstrap_rplugins_io()
    yield bootstrap_result(result)


@prog
@do(NS[CrmRibosome, None])
def post_setup() -> Do:
    yield activate_newly_installed()


@prog.do(None)
def setup_plugins() -> Do:
    yield bootstrap_rplugins()
    yield install_rplugins()
    yield post_setup()


@prog
@do(NS[CrmRibosome, Boolean])
def add_plugins(rplugins: List[Rplugin]) -> None:
    yield Ribo.zoom_main(NS.modify(lambda s: s.append.rplugins(rplugins)))
    yield Ribo.setting(autostart)


@prog.do(None)
def plugins_added(plugins: List[Rplugin]) -> Do:
    autostart = yield add_plugins(plugins)
    if autostart:
        yield setup_plugins()


@prog.do(None)
def init() -> Do:
    plugins = yield initialize()
    yield plugins_added(plugins)


@prog.echo
@do(NS[CrmRibosome, Echo])
def show_plugins() -> Do:
    dir = yield Ribo.setting(venv_dir)
    rplugins = yield Ribo.inspect_main(lambda a: a.rplugins)
    return Echo.info(resources.show_plugins(dir, rplugins))


class AddPluginOptions(Dat['AddPluginOptions']):

    @staticmethod
    def cons(
            name: str=None,
            pythonpath: List[str]=None,
            debug: bool=None,
            interpreter: str=None,
            extensions: List[str]=None,
            track: bool=True,
    ) -> 'AddPluginOptions':
        return AddPluginOptions(
            Maybe.optional(name),
            Maybe.optional(pythonpath),
            Maybe.optional(debug),
            Maybe.optional(interpreter),
            Maybe.optional(extensions),
            Maybe.optional(track),
        )

    def __init__(
            self,
            name: Maybe[str],
            pythonpath: Maybe[List[str]],
            debug: Maybe[bool],
            interpreter: Maybe[str],
            extensions: Maybe[List[str]],
            track: Maybe[bool],
    ) -> None:
        self.name = name
        self.pythonpath = pythonpath
        self.debug = debug
        self.interpreter = interpreter
        self.extensions = extensions
        self.track = track


@prog.do(None)
def add_plugin(spec: str, options: AddPluginOptions) -> Do:
    plugin = cons_rplugin(ConfigRplugin(spec, options.name, options.debug, options.pythonpath, options.interpreter,
                                        options.extensions, options.track))
    yield plugins_added(List(plugin))


__all__ = ('init', 'show_plugins')



import shutil

from amino import do, Path, Do, Lists, env, Maybe, Nothing, IO, List

from ribosome.nvim.io.compute import NvimIO
from ribosome.nvim.io.api import N

from chromatin.settings import interpreter


@do(IO[Path])
def virtualenv_interpreter(venv: str, executable: str) -> Do:
    path_s = yield IO.from_either(env.get('PATH'))
    path = Lists.split(path_s, ':')
    clean_path = path.filter_not(lambda a: a.startswith(venv))
    candidate = yield IO.delay(shutil.which, executable, path=clean_path.mk_string(':'))
    return Path(candidate)


@do(IO[Path])
def find_interpreter(spec: str) -> Do:
    path = Path(spec)
    exists = yield IO.delay(path.exists)
    venv = env.get('VIRTUAL_ENV').get_or_strict('[no venv]')
    yield (
        IO.pure(path)
        if exists else
        virtualenv_interpreter(venv, spec)
    )


def python_interpreter(global_spec: Maybe[str], local_spec: Maybe[str]) -> IO[Path]:
    spec = local_spec.get_or(global_spec.get_or_strict, 'python3.7')
    return find_interpreter(spec)


@do(NvimIO[Path])
def global_interpreter() -> Do:
    global_spec = yield interpreter.value
    yield N.from_io(python_interpreter(global_spec.to_maybe, Nothing))


def join_pythonpath(paths: List[str]) -> str:
    return paths.mk_string(':')


@do(IO[Path])
def which(name: str) -> Do:
    exe = yield IO.delay(shutil.which, name)
    return Path(exe)


def cabal_exe() -> IO[Path]:
    return which('cabal')


def stack_exe() -> IO[Path]:
    return which('stack')


__all__ = ('global_interpreter', 'join_pythonpath', 'stack_exe', 'cabal_exe',)



from amino import Path, List, _
from amino.options import EnvOption
from amino.util.string import plural_s

from chromatin.model.rplugin import Rplugin

xdg_cache_home = EnvOption('XDG_CACHE_HOME')


def create_venv_dir_error(dir: Path) -> str:
    return f'could not create venv dir `{dir}`'


def no_plugins_match_for_update(names: List[str]) -> str:
    return f'no plugins match for update: {names.join_comma}'


def no_plugins_match_for_activation(names: List[str]) -> str:
    return f'no plugins match for activation: {names.join_comma}'


def no_plugins_match_for_deactivation(names: List[str]) -> str:
    return f'no plugins match for deactivation: {names.join_comma}'


def plugins_install_failed(names: List[str]) -> str:
    return f'failed to install rplugins: {names.join_comma}'


def handled_plugins(action: str, names: List[str]) -> str:
    return f'{action} plugin{plural_s(names)}: {names.join_comma}'


def installing_plugins(names: List[str]) -> str:
    msg = handled_plugins('installing', names)
    return f'{msg}…'


def installing_plugin(name: str) -> str:
    return installing_plugins(List(name))


def installed_plugins(names: List[str]) -> str:
    return handled_plugins('installed', names)


def installed_plugin(name: str) -> str:
    return installed_plugins(List(name))


def updated_plugins(names: List[str]) -> str:
    return handled_plugins('updated', names)


def updated_plugin(name: str) -> str:
    return updated_plugins(List(name))


def already_active(names: List[str]) -> str:
    return f'plugin{plural_s(names)} already active: {names.join_comma}'


def show_plugins(venv_dir: Path, plugins: List[Rplugin]) -> str:
    venv_dir_msg = f'virtualenv dir: {venv_dir}'
    plugins_desc = plugins.map(_.spec).cons('Configured plugins:')
    return plugins_desc.cons(venv_dir_msg).join_lines


__all__ = ('xdg_cache_home', 'create_venv_dir_error', 'installed_plugin', 'updated_plugin',
           'no_plugins_match_for_activation', 'no_plugins_match_for_deactivation', 'plugins_install_failed',
           'installed_plugins', 'updated_plugins', 'already_active', 'show_plugins_message', 'installing_plugins',
           'installing_plugin',)



from typing import Any

import sys
import traceback


def echo(nvim: Any, msg: str) -> None:
    safe = msg.replace('"', '\\"')
    nvim.command(f'echo "{safe}"')


def tmpfile_error(msg: str) -> None:
    try:
        import tempfile
        from pathlib import Path
        (fh, file) = tempfile.mkstemp(prefix='chromatin-bootstrap')
        Path(file).write_text(str(msg))
    except Exception as e:
        pass


def run() -> int:
    try:
        from amino.logging import amino_root_logger, amino_root_file_logging
        amino_root_file_logging()
        amino_root_logger.debug('starting chromatin, stage amino')
        try:
            from ribosome.rpc.start import start_external
            from ribosome.rpc.io.start import cons_asyncio_stdio
            from ribosome.logging import nvim_logging
            asio, rpc_comm = cons_asyncio_stdio()
            nvim_api = start_external('define_handlers', rpc_comm).attempt.get_or_raise()
            nvim_logging(nvim_api)
            amino_root_logger.debug('starting chromatin, stage ribosome')
            return stage2(nvim_api)
        except Exception as e:
            amino_root_logger.caught_exception_error(f'importing ribosome in chromatin bootstrap', e)
            raise
    except Exception:
        try:
            msg = traceback.format_exc()
            print(msg, file=sys.stderr)
            tmpfile_error(msg)
        except Exception:
            pass
        return 3


def stage2(nvim: Any) -> int:
    try:
        from amino import Path, Lists, do, Do
        from ribosome.logging import ribo_log
        from ribosome.nvim.io.compute import NvimIO
        from ribosome.nvim.io.data import NSuccess
        from ribosome.nvim.api.exists import wait_for_command
        import chromatin
        from chromatin.host import start_python_host
        ex, bp, ins = Lists.wrap(sys.argv).lift_all(1, 2, 3).get_or_fail(f'invalid arg count for `crm_run`: {sys.argv}')
        python_exe = Path(ex)
        bin_path = Path(bp)
        installed = ins == '1'
        ribo_log.debug(f'starting chromatin, stage 2: {sys.argv}')
        plugin_path = chromatin.__file__
        def error(a: str) -> None:
            raise Exception(f'failed to initialize chromatin: {a}')
        @do(NvimIO[int])
        def run() -> Do:
            channel, pid = yield start_python_host(python_exe, bin_path, plugin_path)
            ribo_log.debug(f'starting chromatin, host started: {channel}/{pid}')
            yield wait_for_command('ChromatinPoll')
            if installed:
                ribo_log.info('chromatin initialized. installing plugins...')
        result = run().run_a(nvim)
        return 0 if isinstance(result, NSuccess) else error(result)
    except Exception as e:
        ribo_log.caught_exception_error('initializing chromatin', e)
        raise e


__all__ = ('run',)



from amino import do, Do, Either, List, Path

from ribosome.nvim.io.state import NS

from chromatin.model.venv import Venv, VenvMeta
from chromatin.env import Env
from chromatin.model.rplugin import Rplugin, InstallableRplugin
from chromatin.venv import cons_venv_under
from chromatin.settings import venv_dir
from chromatin.rplugin import cons_installable_rplugin


@do(NS[Env, List[Rplugin]])
def rplugins_with_crm() -> Do:
    rplugins = yield NS.inspect(lambda a: a.rplugins)
    crm = yield NS.inspect(lambda a: a.chromatin_rplugin)
    return rplugins.cons_m(crm)


@do(NS[Env, Either[str, Rplugin]])
def rplugin_by_name(name: str) -> Do:
    rplugins = yield rplugins_with_crm()
    return rplugins.find(lambda a: a.name == name).to_either(f'no rplugin with name `{name}`')


@do(NS[Env, Venv])
def venv_from_meta(venv_dir: Path, meta: VenvMeta) -> Do:
    rplugin_e = yield rplugin_by_name(meta.name)
    rplugin = yield NS.from_either(rplugin_e)
    return cons_venv_under(venv_dir, rplugin.name)


@do(NS[Env, InstallableRplugin])
def installable_rplugin_from_name(dir: Path, name: str) -> Do:
    rplugin_e = yield rplugin_by_name(name)
    rplugin = yield NS.e(rplugin_e)
    yield NS.m(cons_installable_rplugin.match(rplugin), f'invalid rplugin registered as venv: {rplugin}')


@do(NS[Env, List[InstallableRplugin]])
def installable_rplugins_from_names(names: List[str]) -> Do:
    dir = yield NS.lift(venv_dir.value)
    yield names.traverse(lambda a: installable_rplugin_from_name(dir, a), NS)


__all__ = ('rplugin_by_name', 'venv_from_meta', 'installable_rplugin_from_name', 'installable_rplugins_from_names',)



from typing import Any

from ribosome.rpc.define import ActiveRpcTrigger

from amino import List, Either, Map, Maybe, Nil, L, do, Do
from amino.dat import Dat

from chromatin.model.venv import Venv, VenvMeta
from chromatin.model.rplugin import Rplugin, ActiveRplugin, cons_rplugin, ActiveRpluginMeta


class Env(Dat['Env']):

    @staticmethod
    def cons(
            rplugins: List[Rplugin]=Nil,
            chromatin_rplugin: Rplugin=None,
            chromatin_venv: str=None,
            venvs: List[str]=Nil,
            ready: List[str]=Nil,
            active: List[ActiveRpluginMeta]=Nil,
            uninitialized: List[ActiveRpluginMeta]=Nil,
            triggers: Map[str, List[ActiveRpcTrigger]]=Map(),
            errors: List[str]=Nil,
    ) -> 'Env':
        return Env(
            rplugins,
            Maybe.optional(chromatin_rplugin),
            Maybe.optional(chromatin_venv),
            venvs,
            ready,
            active,
            uninitialized,
            triggers,
            errors,
        )

    def __init__(
            self,
            rplugins: List[Rplugin],
            chromatin_rplugin: Maybe[Rplugin],
            chromatin_venv: Maybe[str],
            venvs: List[str],
            ready: List[str],
            active: List[ActiveRpluginMeta],
            uninitialized: List[ActiveRpluginMeta],
            triggers: Map[str, List[ActiveRpcTrigger]],
            errors: List[str],
    ) -> None:
        self.rplugins = rplugins
        self.chromatin_rplugin = chromatin_rplugin
        self.chromatin_venv = chromatin_venv
        self.venvs = venvs
        self.ready = ready
        self.active = active
        self.uninitialized = uninitialized
        self.triggers = triggers
        self.errors = errors

    def add_plugin(self, name: str, spec: str) -> 'Env':
        return self.append1.rplugins(cons_rplugin(name, spec))

    def add_installed(self, rplugin: str) -> 'Env':
        return self.copy(ready=self.ready.cat(rplugin).distinct)

    @property
    def _str_extra(self) -> List[Any]:
        return List(self.rplugins, self.venvs)

    def rplugin(self, name: str) -> Either[str, Rplugin]:
        return self.rplugins.find(lambda a: a.name == name).to_either(f'no rplugin named `{name}`')

    @property
    def active_packages(self) -> Either[str, List[Rplugin]]:
        return self.active_rplugins.map(lambda a: a.map(lambda a: a.rplugin))

    @property
    def active_package_names(self) -> Either[str, List[str]]:
        return self.active_rplugins.map(lambda a: a.map(lambda a: a.name))

    @property
    def ready_rplugins(self) -> Either[str, List[Rplugin]]:
        return self.ready.traverse(self.rplugin, Either)

    @property
    def inactive(self) -> List[Rplugin]:
        return self.ready.remove_all(self.active_package_names).traverse(self.rplugin, Either)

    def host_started(self, rplugin: ActiveRpluginMeta) -> 'Env':
        return self.append1.active(rplugin).append1.uninitialized(rplugin)

    def initialization_complete(self, failed: List[str]) -> 'Env':
        return self.set.uninitialized(List())

    def deactivate_rplugin(self, meta: ActiveRpluginMeta) -> 'Env':
        return self.mod.active(lambda a: a.without(meta))

    def ready_by_name(self, names: List[str]) -> Either[str, List[Rplugin]]:
        return names.filter(self.ready.contains).traverse(self.rplugin, Either)

    def active_by_name(self, names: List[str]) -> List[ActiveRpluginMeta]:
        return self.active.filter(lambda v: v.rplugin in names)

    def active_rplugin(self, meta: ActiveRpluginMeta) -> Either[str, ActiveRplugin]:
        return (
            self.rplugins
            .find(lambda a: a.name == meta.rplugin)
            .map(lambda p: ActiveRplugin(p, meta))
            .to_either(f'no rplugin for {meta}')
        )

    def as_active_rplugins(self, meta: List[ActiveRpluginMeta]) -> Either[str, List[ActiveRplugin]]:
        return meta.traverse(self.active_rplugin, Either)

    @property
    def active_rplugins(self) -> Either[str, List[ActiveRplugin]]:
        return self.as_active_rplugins(self.active)

    def active_rplugins_by_name(self, names: List[str]) -> Either[str, List[ActiveRplugin]]:
        return self.as_active_rplugins(self.active_by_name(names))

    @property
    def uninitialized_rplugins(self) -> Either[str, List[ActiveRplugin]]:
        return self.as_active_rplugins(self.uninitialized)

    def add_handlers(self, venv: Venv, triggers: List[ActiveRpcTrigger]) -> 'Env':
        return self.append.triggers((venv.name, triggers))

    def handlers_for(self, plugin: str) -> Either[str, List[ActiveRpcTrigger]]:
        return self.triggers.lift(plugin).to_either(f'no triggers defined for {plugin}')

    @property
    def installed_with_crm(self) -> List[VenvMeta]:
        return self.venvs.filter(self.ready.contains).cons_m(self.chromatin_venv)


__all__ = ('Env',)



from amino.case import Case
from amino import List, Path

from chromatin.model.rplugin import VenvRpluginMeta, DistVenvRplugin, DirVenvRplugin


class install_venv_rplugin_args(Case[VenvRpluginMeta, List[str]], alg=VenvRpluginMeta):

    def dist(self, rplugin: DistVenvRplugin) -> List[str]:
        return List(rplugin.req)

    def dir(self, meta: DirVenvRplugin) -> List[str]:
        req = Path(meta.dir) / 'requirements.txt'
        return List('-r', str(req))


__all__ = ('install_venv_rplugin_args',)



from ribosome.config.resources import Resources

from chromatin.config.component import ChromatinComponent
from chromatin.env import Env

ChromatinResources = Resources[Env, ChromatinComponent]

__all__ = ('ChromatinResources',)



from ribosome.data.plugin_state import PluginState

from chromatin.env import Env
from chromatin.config.component import ChromatinComponent

ChromatinState = PluginState[Env, ChromatinComponent]

__all__ = ('ChromatinState',)



from amino.case import Case
from amino import List, do, Do

from ribosome.nvim.io.compute import NvimIO
from ribosome.process import Subprocess
from ribosome.nvim.io.api import N

from chromatin.model.rplugin import (HsRpluginMeta, InstallableRplugin, HsStackageRplugin, HsStackDirRplugin,
                                     HsHackageRplugin)
from chromatin.util.interpreter import stack_exe, cabal_exe


class install_hs_rplugin(Case[HsRpluginMeta, NvimIO[Subprocess[str]]], alg=HsRpluginMeta):

    def __init__(self, rplugin: InstallableRplugin) -> None:
        self.rplugin = rplugin

    @do(NvimIO[Subprocess[str]])
    def hackage(self, a: HsHackageRplugin) -> Do:
        cabal = yield N.from_io(cabal_exe())
        return Subprocess(cabal, List('install', a.dep), self.rplugin.rplugin.name, 600, env=None)

    @do(NvimIO[Subprocess[str]])
    def stackage(self, a: HsStackageRplugin) -> Do:
        stack = yield N.from_io(stack_exe())
        return Subprocess(stack, List('install', a.dep), self.rplugin.rplugin.name, 600, env=None)

    @do(NvimIO[Subprocess[str]])
    def dir(self, a: HsStackDirRplugin) -> Do:
        stack = yield N.from_io(stack_exe())
        return Subprocess(stack, List('install'), self.rplugin.rplugin.name, 600, cwd=a.dir, env=None)


__all__ = ('install_hs_rplugin',)



from typing import Callable, TypeVar

from amino import Path, Try, Right, Either, do, Nil
from amino.do import Do
from amino.boolean import true, false

from ribosome.config.settings import bool_setting, path_setting, list_setting
from ribosome.config.setting import Setting
from ribosome.nvim.io.state import NS
from ribosome.config.resources import Resources

from chromatin.util.resources import xdg_cache_home, create_venv_dir_error

A = TypeVar('A')
D = TypeVar('D')

handle_crm_help = '''When updating plugins, chromatin can also update itself. To take effect, neovim has to be
restarted.
'''
venv_dir_help = '''Chromatin stores installed plugins in separate virtualenvs located in subdirectories of this setting.
'''
rplugin_help = '''The remote plugins managed by chromatin can be configured as a list of dictionaries of the form:
{
    'name': 'title',
    'spec': 'title==0.1.1',
}
`spec` can be any specifier that `pip` understands. If omitted, `name` is used.
'''
autostart_help = '''When set, chromatin will commence installation and activation of plugins on start.
'''
debug_pythonpath_help = '''When set, rplugins will bei started with the outer python env, effectively keeping the
$PYTHONPATH env var as it is in neovim.
'''
autoreboot_help = '''When set, plugins will be de- and reactivated after being updated.
'''
interpreter_help = '''Overrides the python executable that is used to create virtualenvs. If unset, the interpreter is
determined heuristically. This can be nontrivial, for example when already inside a virtualenv (mostly relevant for
development), since creating a venv from within another doesn't work correctly.
'''


@do(Either[str, Path])
def default_venv_dir() -> Do:
    xdg_cache_path = xdg_cache_home.value / Path | (Path.home() / '.cache')
    venv_dir = xdg_cache_path / 'chromatin' / 'venvs'
    yield Try(venv_dir.mkdir, parents=True, exist_ok=True).lmap(lambda a: create_venv_dir_error(venv_dir))
    yield Right(venv_dir)


# FIXME validate `rplugins` correctly; if it is e.g. a list[str], this will result in an exception in `read_conf`

handle_crm = bool_setting('handle_crm', 'update chromatin', handle_crm_help, True, Right(true))
venv_dir = path_setting('venv_dir', 'virtualenv base directory', venv_dir_help, True, default_venv_dir())
rplugins = list_setting('rplugins', 'rplugin config', rplugin_help, True, Right(Nil))
autostart = bool_setting('autostart', 'autostart plugins', autostart_help, True, Right(true))
debug_pythonpath = bool_setting('debug_pythonpath', 'pass through $PYTHONPATH', debug_pythonpath_help, True,
                                Right(false))
autoreboot = bool_setting('autoreboot', 'autoreboot plugins', autoreboot_help, True, Right(true))
interpreter = path_setting('interpreter', 'python interpreter for venvs', interpreter_help, True)


@do(NS[D, None])
def ensure_setting(setting: Setting[A], value: A) -> Do:
    yield NS.lift(setting.ensure(value))


__all__ = ('ChromatinSettings', 'setting', 'update_setting', 'ensure_setting', 'handle_crm', 'venv_dir', 'rplugins',
           'autostart', 'debug_pythonpath', 'autoreboot', 'interpreter',)



from amino import Dat

from ribosome.config.component import Component, ComponentData

from chromatin.env import Env


class ChromatinComponent(Dat['ChromatinComponent']):

    @staticmethod
    def cons() -> 'ChromatinComponent':
        return ChromatinComponent()

    def __init__(
            self,
    ) -> None:
        pass


Comp = Component[ComponentData, ChromatinComponent]

__all__ = ('ChromatinComponent',)



import abc
import pkg_resources

from amino import Path, Boolean
from amino.boolean import true, false
from amino.dat import ADT, Dat

from chromatin.model.rplugin import Rplugin


class VenvMeta(Dat['VenvMeta']):

    def __init__(
            self,
            rplugin: str,
            dir: Path,
            python_executable: Path,
            bin_path: Path,
    ) -> None:
        self.rplugin = rplugin
        self.dir = dir
        self.python_executable = python_executable
        self.bin_path = bin_path

    @property
    def name(self) -> str:
        return self.rplugin


class Venv(Dat['Venv']):

    def __init__(self, name: str, meta: VenvMeta) -> None:
        self.name = name
        self.meta = meta


class VenvStatus(ADT['VenvStatus']):

    @abc.abstractproperty
    def exists(self) -> Boolean:
        ...


class VenvPresent(VenvStatus):

    def __init__(self, plugin: Rplugin, venv: Venv) -> None:
        self.plugin = plugin
        self.venv = venv

    @property
    def exists(self) -> Boolean:
        return true


class VenvAbsent(VenvStatus):

    def __init__(self, rplugin: Rplugin) -> None:
        self.rplugin = rplugin

    @property
    def exists(self) -> Boolean:
        return false


class VenvPackageStatus(ADT['VenvPackageStatus']):

    @abc.abstractproperty
    def exists(self) -> Boolean:
        ...


class VenvPackageExistent(VenvPackageStatus):

    def __init__(self, venv: Venv, dist: pkg_resources.Distribution) -> None:
        self.venv = venv
        self.dist = dist

    @property
    def exists(self) -> Boolean:
        return true


class VenvPackageAbsent(VenvPackageStatus):

    def __init__(self, venv: Venv) -> None:
        self.venv = venv

    @property
    def exists(self) -> Boolean:
        return false


__all__ = ('VenvStatus', 'VenvPresent', 'VenvAbsent', 'VenvPackageAbsent', 'VenvPackageExistent', 'VenvPackageStatus',
           'Venv',)
